from typing import Any, Dict, List
from haystack import component, default_from_dict, default_to_dict, Document

from creao.core.Dedup import Dedup

@component
class Deduplication:
    def __init__(self):
        pass

    def warm_up(self):
        """
        Initializes the component.
        """
        if not hasattr(self, "dedup"):
            self.dedup = Dedup()

    @component.output_types(documents=List[Document])
    def run(self, documents: List[Document]):
        if not hasattr(self, "dedup"):
            raise RuntimeError("The embedding model has not been loaded. Please call warm_up() before running.")
        if len(documents) == 0:
            return {"dedup_list":[]}
        file_name = documents[0].meta["file_name"]
        chunk = documents[0].meta["chunk"]
        input_texts = [doc.content for doc in documents]
        res = self.dedup.execute(input_texts)
        dedup_list = []
        for item in res:
            doc = Document(meta={"file_name":file_name, "chunk":chunk}, content=item)
            dedup_list.append(doc)
        return {"documents":dedup_list}
    
    
    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(
            self,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Dedup":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)


from typing import Any, Dict, List
from haystack import component, default_from_dict, default_to_dict, Document
from creao.core.Endpoints import OpenAILLM, CreaoLLM
import json
from pydantic import BaseModel


extract_user_interest_prompt = """\
You are given a Persona and a Passage. Your task is to immitate the persona and create a list interesting topics from the given passage.

<Persona>
{persona}
</Persona>

<Passage>
The following information is from a file with the title "{file_name}".

{passage}
</Passage>

Answer format - Generate a json with the following fields
- "list_of_interest": [<fill with 1-5 word desription>]

Use Reflective Thinking: Step back from the problem, take the time for introspection and self-reflection. Examine personal biases, assumptions, and mental models that may influence problem-solving, and being open to learning from past experiences to improve future approaches. Show your thinking before giving an answer.
"""

class POISchema(BaseModel):
    list_of_interest: list[str]

@component
class ExtractPOI:
    def __init__(self, custom_prompt: str = None, service: str = "default"):
        self.custom_prompt = custom_prompt
        self.service = service
        if self.service == "default":
            self.llm = CreaoLLM(bot_name="point of interests extraction assistant", bot_content="You are given a Persona and a Passage. Your task is to adopt the given Persona and reflect deeply on the content of the Passage. Then, create a list of interesting topics from the Passage that align with the Persona's unique perspective and thinking style.Reflective Thinking: Before you generate the final list of topics, pause to examine your assumptions, biases, and the mental models that the Persona might use. Consider how the Persona's perspective influences what they find interesting, and how they prioritize topics. Be open to learning from previous similar tasks and improving the outcome. Explicitly share your reflective thought process before giving the final answer.")
        elif self.service == "openai":
            self.llm = OpenAILLM()

    @component.output_types(documents=List[Document])
    def run(self, personas:List[str], file_name:str, chunk:str):
        #print("extracting points of interest with minimax")
        res_list = []
        for persona in personas:
            prompt = extract_user_interest_prompt.format(
                persona=persona, file_name=file_name, passage=chunk
            )
            if self.service == "default":
                response_json = self.llm.invoke(prompt,{
                        "list_of_interest": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                        }
                    },"ExtractPOI")
                try:
                    raw_answer = json.loads(response_json["reply"])
                except Exception as e:
                    print(f"ExtractPOI json decode error:{e}, response_json:{response_json}")
                    raw_answer = {"list_of_interest":[]}
            elif self.service == "openai":
                try:
                    raw_answer = json.loads(self.llm.invoke(prompt, POISchema))
                except Exception as e:
                    print(f"ExtractPOI json decode error:{e}")
                    raw_answer = {"list_of_interest":[]}
            for interest in raw_answer["list_of_interest"]:
                doc = Document(meta={"persona":persona, "file_name":file_name, "chunk":chunk}, content=interest)
                res_list.append(doc)
        #print("extracting points of interest with minimax done")
        return {"documents":res_list}
        
    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(
            self,
            custom_prompt=self.custom_prompt,
            service=self.service
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExtractPOI":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)


from openai import OpenAI
from haystack.components.generators import OpenAIGenerator
from haystack.utils import Secret
import time
import requests
import os

class Embed:
    def __init__(self):
        self.client = OpenAI()

    def invoke(self, text):
        response = self.client.embeddings.create(
            model="text-embedding-3-small",
            input=text,
        )
        return response.data[0].embedding
    

class CreaoLLM:
    def __init__(self, bot_name="assistant", bot_content="assistant") -> None:
        self.bot_name = bot_name
        self.bot_content = bot_content
    
    def invoke(self, prompt, schema, component_id="default"):
        """
        Invoke the Creao LLM API
        """
        # The API Gateway endpoint URL
        url = 'https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver'
        api_key= os.environ["CREAO_API_KEY"]
        headers = {"Content-Type":"application/json", "x-api-key":api_key}
        # Payload to be sent to the Lambda function via API Gateway
        payload = {
            "prompt": prompt,
            "schema": schema,
            "bot_name": self.bot_name,
            "bot_content": self.bot_content,
            "component_id": component_id
        }
        # Send the request
        response = requests.post(url,headers=headers, json=payload)
        #print("reponse:", response.content)
        try:
            return response.json()
        except Exception as e:
            print(f"CreaoLLM json decode error:{e}, with response:{response.text}")
            return {"reply":""}

class MiniMaxPro:
    def __init__(self, bot_name="assistant", bot_content="assistant"):
        self.bot_name = bot_name
        self.bot_content = bot_content
        group_id = "1820662545770364967" 
        api_key = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9.eyJHcm91cE5hbWUiOiJDcmVhbyBBSSIsIlVzZXJOYW1lIjoiQ3JlYW8gQUkiLCJBY2NvdW50IjoiIiwiU3ViamVjdElEIjoiMTgyMDY2MjU0NTc3ODc1MzU3NSIsIlBob25lIjoiMTg2Njc4OTMzOTkiLCJHcm91cElEIjoiMTgyMDY2MjU0NTc3MDM2NDk2NyIsIlBhZ2VOYW1lIjoiIiwiTWFpbCI6ImthaUBjcmVhby5haSIsIkNyZWF0ZVRpbWUiOiIyMDI0LTA4LTIwIDAzOjM4OjQ3IiwiaXNzIjoibWluaW1heCJ9.PUSbHFa6t9Po2z8AwbVlJIgWm7-zRL9047AB_RBN8azCQxNzErJE_6wd8u4726xuIS7NZ6768LJW42-etD8i6N7Dza92QsGkAmEg3PHaYAzbOQgQVge0G_OCpJLMnVOCpZi7yOlkPNuu6BDW8-6lOtJoAeNkatxWJqL-vp_O_zyqgUwoz-2DGiY9H7pQI0Z0KGkj832RG05Lg6T-CJZKKVynT4cwnWqVTX4jYidCmsn6vi45MqCrB78o7v1jJtwdCui74iTRY8n23Qxp3Jis6F0cfFY2kIn_OhXBMb2p6QwEOgfxJRzv1zyusf3MTYTg7aqo37Lnda3uWRC34QSGJA"
        url = f"https://api.minimax.chat/v1/text/chatcompletion_pro?GroupId={group_id}"
        headers = {"Content-Type":"application/json", "Authorization":"Bearer " + api_key}
        self.headers = headers
        self.url = url
    def invoke(self, prompt, schema):
        payload = {
            "model": "abab6.5s-chat",
            "tokens_to_generate": 2048,
            "temperature": 0.1,
            "top_p": 0.95,
            "stream": False,
            "reply_constraints": {
                "sender_type": "BOT",
                "sender_name": self.bot_name,
                "glyph": {
                    "type": "json_value",
                    "json_properties": schema
                }
            },
            "sample_messages": [],
            "plugins": [],
            "messages": [
                {
                "sender_type": "USER",
                "sender_name": "用户",
                "text": prompt
                }
            ],
            "bot_setting": [
                {
                "bot_name": self.bot_name,
                "content": self.bot_content
                }
            ]
            }
        i = 0
        status = -1
        while status != 0:
            response = requests.post(self.url, headers=self.headers, json=payload)
            status = response.json()["base_resp"]["status_code"]
            if status == 2045: # hit rate limit
                time.sleep(3)
            i += 1
            if i > 5:
                break
        return response.json()


class OpenAILLM:
    def __init__(self, model_id = "gpt-4o-mini"):
        self.client = OpenAI()
        self.model_id = model_id

    def invoke(self, prompt, schema=None):
        messages = [{"role": "user", "content": prompt}]
        try:
            if schema is None:
                response = self.client.chat.completions.create(
                    model=self.model_id,
                    messages=messages,
                    temperature=0,
                    max_tokens=1024,
                )
                return response.choices[0].message.content
            else:
                response = self.client.beta.chat.completions.parse(
                    model=self.model_id,
                    messages=messages,
                    temperature=0,
                    max_tokens=1024,
                    response_format=schema,
                )
            return response.choices[0].message.content
        except Exception as e:
            print(e)
            return None
        
class MiniMax:
    def __init__(self):
        print("initialize minimax")
        self.client = OpenAIGenerator(
            api_base_url="https://api.minimax.chat/v1",
            api_key=Secret.from_env_var("MINIMAX_API_KEY"),
            model="abab6.5s-chat"
        )

    def invoke(self, prompt):
        response = self.client.run(prompt)
        return response["replies"]
    

#creaoLM = CreaoLLM()
#r = creaoLM.invoke("hello", {"name":"world"})
#print(r)


from typing import Any, Dict, List, Optional
from creao.core.Endpoints import CreaoLLM
from datasets import Dataset
from sentence_transformers.evaluation import (
    InformationRetrievalEvaluator,
    SequentialEvaluator,
)
from datasets import concatenate_datasets

from sentence_transformers.util import cos_sim
from creao.core.Diagnosis import RetrieverDiagnoser
from creao.core.demo.rag import RAG
from ragas.metrics import answer_relevancy
from ragas.metrics import faithfulness

from ragas import evaluate as ragas_evaluate
from langchain_core.language_models.llms import LLM
from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from datasets import load_dataset

class CreaoLangchainEmbedding:
    def __init__(self) -> None:
        self.embedding_model = CreaoLLM()

    def embed_query(self, text: str) -> List[float]:
        return self.embedding_model.invoke(text,"",component_id="embed")

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        return [self.embedding_model.invoke(t,"",component_id="embed") for t in texts]

class CreaoLangchainLLM(LLM):
        
    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        creaollm = CreaoLLM()
        response_raw = creaollm.invoke(prompt, "", "ragas_custom_llm")
        return response_raw["reply"]
 
    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters."""
        return {
            # The model name allows users to specify custom token counting
            # rules in LLM monitoring applications (e.g., in LangSmith users
            # can provide per token pricing for their model and monitor
            # costs for the given LLM.)
            "model_name": "CreaoLangchainLLM",
        }

    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model. Used for logging purposes only."""
        return "creao_langchain_llm"

class RagDiagnoser(object):
    """
    The RagDiagnoser class.
    
    # Load the dataset
    dataset = load_dataset("your_dataset")

    # Create an instance of the RagDiagnoser class
    diagnoser = RagDiagnoser(dataset)

    # Generate a diagnosis report
    report = diagnoser.generate_rag_diagnosis_report(full_dataset=True, question_limit=100)

    # Extract retrieval errors
    errors = diagnoser.extract_retrieval_errors(rank_threshold=2)

    # Create a RAG chatbot demo
    diagnoser.create_rag_chatbot_demo()

    # Get all unique chunks
    chunks = diagnoser.get_all_unique_chunks()

    # Construct an information retrieval evaluator
    evaluator = diagnoser.construct_information_retireval_evaluator()
    """
    def __init__(self, dataset: Dataset, embedding_model_id:str="BAAI/bge-base-en-v1.5", service="default") -> None:
        """
        Initialize the RagDiagnoser.
        :param dataset: The dataset to use for evaluation.
        :param embedding_model_id: The embedding model id to use.
        :param service: The service to use for the LLM.
        """
        self.service = service
        self.train_dataset = dataset["train"]
        self.test_dataset = dataset["test"]
        chunks = set()
        for item in self.test_dataset:
            chunks.add(item["positive"])
        for item in self.train_dataset:
            chunks.add(item["positive"])
        self.chunks = chunks
        print(f"Number of unique chunks: {len(chunks)}")
        question_list = []
        if service == "default":
            self.creao_langchain_llm = CreaoLangchainLLM()
            self.creao_langchain_embedding = CreaoLangchainEmbedding()
        for item in self.test_dataset:
            question_list.append({"question":item["anchor"], "positive":item["positive"]})
        retriever_diagnoser = RetrieverDiagnoser(self.chunks, question_list, embedding_model_id=embedding_model_id)
        self.retriever_diagnoser = retriever_diagnoser
        self.rag_system = RAG(self.chunks, embedding_model_id=embedding_model_id,service=service)

    def get_rag_system(self):
        return self.rag_system
    
    def create_rag_chatbot_demo(self):
        self.rag_system.generate_gradio_interface()

    def generate_rag_diagnosis_report(self, full_dataset=False, question_limit:int=-1):
        """
        Generate a diagnosis report for the RAG system.
        :param full_dataset: Whether to use the full dataset or not.
        :param question_limit: The number of questions to limit the diagnosis to.
        :return: The diagnosis report.
        """
        def get_positive_context_rank(positive: str, contexts:List[str]):
            rank = -1
            if positive in contexts:
                rank = contexts.index(positive)
            return rank
        def get_context_and_answer(question:str):
            response =self.rag_system.query_response(question)
            answer = response["llm"]["replies"][0]
            context = [doc.content for doc in response["retriever"]["documents"]]
            return context, answer
        if full_dataset:
            question_list = concatenate_datasets([self.train_dataset, self.test_dataset])
        else:
            question_list = self.test_dataset
        if question_limit >= 0:
            question_list = question_list.select(range(question_limit))
        questions = []
        answers = []
        contexts = []
        ids = []
        postives = []
        personas = []
        styles = []
        original_questions = []
        positive_ranks = []
        for item in question_list:
            question = item["anchor"]
            postive = item["positive"]
            persona = item["persona"]
            style = item["style"]
            original_question = item["original_question"]
            id = item["id"]
            context, answer = get_context_and_answer(question)
            positive_rank = get_positive_context_rank(postive, context)
            questions.append(question)
            styles.append(style)
            original_questions.append(original_question)
            postives.append(postive)
            positive_ranks.append(positive_rank)
            personas.append(persona)
            answers.append(answer)
            contexts.append(context)
            ids.append(id)
        data_samples = {
            "positive_ranks": positive_ranks,
            "question": questions,
            "answer": answers,
            "positive": postives,
            "persona": personas,
            "contexts": contexts,
            "id": ids,
            "style": styles,
            "original_question": original_questions
        }
        rag_dataset = Dataset.from_dict(data_samples)
        if self.service == "default":
            score = ragas_evaluate(rag_dataset,llm=self.creao_langchain_llm, embeddings=self.creao_langchain_embedding, metrics=[answer_relevancy])
        elif self.service == "openai":
            score = ragas_evaluate(rag_dataset, metrics=[answer_relevancy])
        return score

    def get_all_unique_chunks(self):
        return self.chunks

    def extract_retrieval_errors(self, rank_threshold:int=2):
        # list of dict with question, and positive chunk (ground truth)
        filtered_res = self.retriever_diagnoser.detect_retriever_error(rank_threshold=rank_threshold)
        return filtered_res


    def construct_information_retireval_evaluator(self):
        matryoshka_dimensions = [768, 512, 256, 128, 64]  # Important: large to small\
        # cheat a chunk to dictionary index
        chunk_to_index = {}
        i = 0
        for item in self.chunks:
            chunk_to_index[item] = i
            i += 1
        # constreuct index to chunk
        index_to_chunk = {v: k for k, v in chunk_to_index.items()}
        # test queries for information retrieval evaluation
        queries = dict(zip(self.test_dataset["id"], self.test_dataset["anchor"]))
        # relevant_docs:  Query ID to relevant documents (qid => set([relevant_cids])
        relevant_docs = {}

        for i in range(len(self.test_dataset)):
            chunk = self.test_dataset[i]["positive"]
            id = self.test_dataset[i]["id"]
            relevant_docs[id] = [chunk_to_index[chunk]]

        matryoshka_evaluators = []
        # Iterate over the different dimensions
        for dim in matryoshka_dimensions:
            ir_evaluator = InformationRetrievalEvaluator(
                queries=queries,
                corpus=index_to_chunk,
                relevant_docs=relevant_docs,
                name=f"dim_{dim}",
                truncate_dim=dim,  # Truncate the embeddings to a certain dimension
                score_functions={"cosine": cos_sim},
                write_csv=True
            )
            matryoshka_evaluators.append(ir_evaluator)

        # Create a sequential evaluator
        evaluator = SequentialEvaluator(matryoshka_evaluators)
        return evaluator
    


